use std::collections::HashMap;
use std::io;

use axum::extract::multipart::{self, Multipart};
use sha2::Digest as _;
use tokio::io::{AsyncWrite, BufWriter};
use tokio_util::io::StreamReader;

use crate::prelude::*;

#[derive(Debug, Deserialize)]
pub struct Params {
    parent_folder: Uuid,
}

async fn write_file(
    file_id: Uuid,
    file: impl AsyncWrite + Unpin,
    file_name: &str,
    field: &mut multipart::Field<'_>,
    parent_folder: Uuid,
    pool: &Pool,
) -> bool {
    const BUF_CAP: usize = 64 * 1024 * 1024; // 64 MiB
    let mut hash = sha2::Sha512::new();
    let mut size: i64 = 0;
    let stream = field.map(|value| match value {
        Ok(bytes) => {
            hash.update(&bytes);
            size = i64::try_from(bytes.len())
                .ok()
                .and_then(|part_size| size.checked_add(part_size))
                .ok_or_else(|| io::Error::other(anyhow::anyhow!("Size calculation overflow")))?;
            Ok(bytes)
        }
        Err(err) => Err(io::Error::other(err)),
    });
    let mut reader = StreamReader::new(stream);
    let mut writer = BufWriter::with_capacity(BUF_CAP, file);
    if let Err(err) = tokio::io::copy(&mut reader, &mut writer).await {
        tracing::warn!(%err);
        return false;
    }
    let hash = hash.finalize().to_vec();
    db::file::insert(file_id, parent_folder, file_name, size, hash, pool)
        .await
        .inspect_err(|err| tracing::warn!(%err))
        .is_ok()
}

pub async fn upload(
    Query(params): Query<Params>,
    State(state): State<AppState>,
    claims: Claims,
    mut multi: Multipart,
) -> Result<Json<HashMap<String, Uuid>>, StatusCode> {
    db::folder::get_permissions(params.parent_folder, claims.user_id, &state.pool)
        .await
        .handle_internal()?
        .can_write_guard()?;

    let existing_names = db::folder::get_names(params.parent_folder, &state.pool)
        .await
        .handle_internal()?;
    let mut result = HashMap::new();
    while let Ok(Some(mut field)) = multi.next_field().await {
        let Some(file_name) = field.file_name().map(ToOwned::to_owned) else {
            continue;
        };
        if existing_names.contains(&file_name) {
            continue;
        }
        let Ok((file_id, mut file)) = state.storage.create().await else {
            tracing::warn!("Couldn't create uuid for new file");
            continue;
        };
        let is_success = write_file(
            file_id,
            &mut file,
            &file_name,
            &mut field,
            params.parent_folder,
            &state.pool,
        )
        .await;
        if !is_success {
            let _ = state.storage.delete(file_id).await;
            continue;
        }
        result.insert(file_name, file_id);
    }

    Ok(Json(result))
}
